//-----------------------------------------------------------------------------
// pilotage d'un ecran VGA pour affichage hyperterminal UART TTL
// L. BAGHLI 08/01/2008
// 30 MIPS, 5 V 
//-----------------------------------------------------------------------------
#include "p30F4012.h"
#include "symbol_tbl.h"	

//Configuration bits
// Q=7.3728 MHz
_FOSC(CSW_FSCM_OFF & XT_PLL16);	// 7.3728MHz *16 = 117.9648 MHz /4 = 29.4912 MIPS maxi pour ce pic
_FWDT(WDT_OFF);
_FBORPOR(PBOR_OFF & BORV_27 & PWRT_16 & MCLR_EN);
_FGS(CODE_PROT_OFF);
//-----------------------------------------------------------------------------
//Program Specific Constants
#define FCY 29491200					//Instruction cycle rate (Osc x PLL / 4)
#define T1Period 938					// pour 31.77 us (31.469 kHz de balayage Vertical)  29.4912 MHz, T1Period = 937=FCY/31469 Hz
#define tcntPRD 31469					// combien de fois pour ariver en 1s avec des pas de 31.77 us : 31469
#define MILLISEC FCY/29491			// 1 mSec delay constant

//// Q=10 MHz
//_FOSC(CSW_FSCM_OFF & XT_PLL8);	//10Mhz *8 = 80 MHz /4 = 20 MIPS maxi pour ce pic
//_FWDT(WDT_OFF);
//_FBORPOR(PBOR_OFF & BORV_27 & PWRT_16 & MCLR_EN & PWMxH_ACT_HI & PWMxL_ACT_LO);
//_FGS(CODE_PROT_OFF);
////-----------------------------------------------------------------------------
////Program Specific Constants
//#define FCY 20000000						//Instruction cycle rate (Osc x PLL / 4) = 20 MIPS
//#define FPWM 16000						//PWM freq
//#define HalfDUTY 	624	// set the pwm half duty = PTPER = 624UL=FCY/FPWM/2-1;
//#define FullDUTY HalfDUTY*2			//  car PDCx a une double resolution % PTPER
//#define MILLISEC FCY/20000			// 1 mSec delay constant

// ressources :
//http://www.circuitdb.com/circuits/id/69
//http://yusoft.kulichki.com/english/pic/my_proj.htm

// VGA_G sortie directement sur la SPI MOSI 
#define VGA_G				LATFbits.LATF3		// Output
#define VGA_debug		LATEbits.LATE0		// Output
#define VGA_HSync		LATEbits.LATE1		// Output
#define VGA_VSync		LATEbits.LATE2		// Output
#define InfoLED			LATDbits.LATD0		// Output

//	SPI1CON = 0x013F;		// Master mode, SCK = Fcy/1/1 = 29.4912 MHz 1|1111), CKP=0 CKE=1 : sampled on the rising edge of SCLK, Clk idle is low, 8 bits
//	SPI1CON = 0x013B;		// Master mode, SCK = Fcy/2/1 = 14.7456 MHz 1|1011), CKP=0 CKE=1 : sampled on the rising edge of SCLK, Clk idle is low, 8 bits
//	SPI1CON = 0x0123;		// Master mode, SCK = Fcy/8/1 = 3.6864 MHz 0|0011), CKP=0 CKE=1 : sampled on the rising edge of SCLK, Clk idle is low, 8 bits
//	SPI1CON = 0x012E;		// Master mode, SCK = Fcy/5/4 = 1.47456 MHz 0|1110), CKP=0,  CKE=1, Clk idle is low, 8 bits
#define SPI1CON_29_5MHz 0x013F
#define SPI1CON_14_7MHz 0x013B

//Makro for video handle
#define video_off		VGA_debug=0				//Video (Green Color)
#define video_on		VGA_debug=1
#define hsync_off		VGA_HSync=1		//HSYNC
#define hsync_on		VGA_HSync=0
#define vsync_off		VGA_VSync=1		//VSYNC
#define vsync_on		VGA_VSync=0

void setup_ports(void);
void DelayNmSec(unsigned int N);
void InitVar();
void InitUART();
void ClearDisplayBuffer();


struct {
			unsigned Running		: 	1;
			unsigned SendData		:		1;
			unsigned unused 		:		14;
		}	Flags;

//Global definitions for VGA render
#define vga_field_line_count 525 //standart VGA quantity lines
//#define vga_symbols_per_row 40   //symbols quantity per horizontal
#define vga_symbols_per_row 45   //symbols quantity per horizontal
#define vga_row_count 40         //symbols quantity per vertical
#define vga_row_countm1 39       //symbols quantity per vertical-1
#define vga_symbol_height 12		 //rendered symbol height

//Global variables
unsigned char str_array[vga_row_count*vga_symbols_per_row]; //Chars array for display buffer
unsigned int current_symbol; //Current symbol number at terminal

unsigned char video_enable_flg;//flag for enable render
unsigned char raw_render;//Current row at display buffer for render
unsigned int linecount;// current line for render
unsigned char y_line_render;	//Current Y-line at symbol for render 

//Store strings data at Program Space, to avoid RAM leakage
const unsigned char Phrase[]="LOTFI is here and brings you the best        VGA terminal display (40 lines x 45 rows)    on dsPIC 30F4012 !";

unsigned int heure, min,sec,tcnt;

// RS232 -------------------------------------------
void InitUART(void);
void SendMsg(void);
#define CR	0x0D
#define LF 	0x0A
#define BAUD 19200
int SeqComm;		// ttes les 0.5 s
#define SeqCommMax 5000

//-----------------------------------------------------------------------------
//	Initialise variables
//-----------------------------------------------------------------------------
void InitVar(void)
{
char * ptr;
unsigned char i;
	linecount = 0;
	ClearDisplayBuffer();
	current_symbol = 0x0;
// debug rempli par une phrase
	ptr = &str_array[20* vga_symbols_per_row];	// 20me ligne
	i=0;
	while( Phrase[i]!=0 )
		{
		*ptr++ = Phrase[i++];
		}

//	str_array[5* vga_symbols_per_row]='I';
}
//-----------------------------------------------------------------------------
//	Setup ports
//-----------------------------------------------------------------------------
void setup_ports(void)
{
	ADPCFG = 0xFFFF;				// all PORTB = Digital(1) ie 1111 1111
	// Clear All Ports Prior to defining I/O
	PORTB=0;	//Initialize LED pin data to off state
	PORTC=0;
	PORTD=0;
	PORTE=0;	
	// Now set pin direction registers
	TRISB = 0xFFFF;		// RB0-5 inputs : NC
	TRISC = 0xDFFF;		// U1ATX/RC13 in , U1ATX/RC14 out , inutile pour le UART donc 0xFFFF is also ok
	TRISD = 0xFFFE;		// InfoLED RD0 out, RD1 in NC
	TRISE = 0x01F8;		// VGA_debug RE0 out, VGA_HSync RE1 out, VGA_VSync RE2 out, RE8 in NC : 1 1111 1000
	TRISF = 0xFFF7;		// SDI1/RF2 in, SDO1/RF3 out : 0111 car je dois forcer le MOSI apres le SPI  0 et non pas le laisser  la dernire valeur
//	TRISF = 0xFFFF;		// RF input NC
	// SPI pour le output shift du G(Green component)
	SPI1STAT = 0x8000;	// Enable SPI port
	SPI1CON = SPI1CON_29_5MHz;		// vitesse max du SPI
//	SPI1CON = SPI1CON_14_7MHz;
//	SPI1CON = 0x013E;		// Master mode, SCK = Fcy/1/4 = 7.3728 MHz 1|1110), CKP=0 CKE=1 : sampled on the rising edge of SCLK, Clk idle is low, 8 bits
}
//-----------------------------------------------------------------------------
void initTimer(void)
{
// Timer1 pour l ISR des 31.777 us
	T1CON = 0;					// ensure Timer 1 is in reset state, internal timer clock Fosc/4, no prescale
	TMR1  = 0;					// RAZ Timer1
 	IFS0bits.T1IF = 0;	// reset Timer 1 interrupt flag 
	IPC0bits.T1IP = 4;	// set Timer1 interrupt priority level to 4 
 	IEC0bits.T1IE = 1;	// enable Timer 1 interrupt
	PR1 = T1Period;			// set Timer 1 period register

	T1CONbits.TON = 1;	// enable Timer 1 and start the count
}
//------------------------------------------------------------------------
void __attribute__((interrupt, auto_psv)) _T1Interrupt( void )
{
unsigned char j;
char * _ptr;
const char * _ptr1;
//	InfoLED=1;	
	IFS0bits.T1IF = 0;	
	hsync_on;		// dbut de la raster
	//Count number of lines
	if (++linecount == vga_field_line_count)
			{
			linecount = 0;
			//clear pointers for render display buffer
			raw_render = 0; 
			y_line_render = 0;
			}
		else
			{
			__builtin_nop();
			__builtin_nop();
			__builtin_nop();
			}
	//Make Vsync length 2 VGA lines 
	if ((linecount == 10 )||(linecount == 11 ))
							vsync_on;			//Make here vertical syncronization & HSYNC syncro level on
			else		vsync_off;		//.. & HSYNC syncro level on

	video_enable_flg = 1;

	if (linecount < 45)
			{
			video_enable_flg = 0;
			//Add to avoid flickering at top display
			__builtin_nop();					
			__builtin_nop();					
			__builtin_nop();					
			__builtin_nop();					
			__builtin_nop();					
			__builtin_nop();					
			}
		else
			{
			 //Forming current string for rendering
			 if (++y_line_render == vga_symbol_height)
					{
					raw_render++;
					y_line_render = 0;
					}
				else
					{
					__builtin_nop();
					}
			}
	hsync_off; //HSYNC syncro level off
	for (j=0; j<28; j++)			__builtin_nop();

// render ici pour tre sr d'tre synchrone
	if(video_enable_flg)
			{
			if (raw_render > vga_row_countm1)		raw_render=vga_row_countm1;
			//on est ds la zone visible, alors : render
			_ptr = &str_array[raw_render * vga_symbols_per_row];		//Set pointer for render line (display bufffer)
			_ptr1 = &symbol[0][y_line_render];				//Set pointer for render line (character generator)
			//Cycle for render line 
			j = vga_symbols_per_row;
			video_on;	// pour dbug
			while(j--)
				{
	//				uidummy = SPI1BUF;
				SPI1BUF = ~(*(_ptr1 + (* _ptr++)*vga_symbol_height));
				//while (SPI1STATbits.SPITBF);
				//while ( !SPI1STATbits.SPIRBF);
				SPI1BUF = 0;	// pour effacer le bit restant, mais il faut aller vite
				}
			video_enable_flg=0;
			video_off;	// pour dbug
			}

//	if (++tcnt>=tcntPRD)	
//						{
//						tcnt=0;
//						if (++sec==60)
//									{
//									sec=0;
//									if (++min==60)
//												{
//												min=0;
//												if (++heure==24)	heure=0;	
//												}
//									}
//						// comme l'heure hh:mm:ss a chang => on doit l'afficher
//						}
// communication dsPIC -> PC
//	if (!--SeqComm)		{
//										SeqComm=SeqCommMax;
//										Flags.SendData=1;
//										}

//	InfoLED=0;
}
//---------------------------------------------------------------------
void ClearDisplayBuffer()								//Clear display buffer
{
unsigned int i = vga_row_count*vga_symbols_per_row;
char * ptr;
	//Set pointers for clear string array
	ptr = &str_array[0];
	while(i--)
		*ptr++ = 0x0;
}
//---------------------------------------------------------------------
// Below are the interrupt vectors for the serial receive and transmit
//---------------------------------------------------------------------
void __attribute__((interrupt, auto_psv)) _U1TXInterrupt(void)
{
	IFS0bits.U1TXIF = 0;	// clear interrupt flag
}
//---------------------------------------------------------------------
void __attribute__((interrupt, auto_psv)) _U1RXInterrupt(void)
{
// Parser received symbols from UART
unsigned int i;
char * ptr;
unsigned char current_line, received_symbol;
	IFS0bits.U1RXIF = 0;	// clear interrupt flag
	received_symbol = U1RXREG;
	//Check for overflow display buffer
	if(current_symbol == (vga_row_count*vga_symbols_per_row))
				{
//				ClearDisplayBuffer();	il n'y a plus d'efface en entier mais ligne par ligne
				current_symbol = 0x0;		// remet au dbut du tableau
//				// vide le buffer car on a pris bcp de tps  effacer...
//				U1STAbits.OERR=0;	// y a surement eu trop de donner alors clear overflow to continue receiving
//				return;
				}
	
	switch ( received_symbol )
		{
		//BackSpace
		case  0x08: 
			if(current_symbol)
				{
				str_array[current_symbol] = 0x0;
				str_array[--current_symbol] = 0x0;
			}
			break;
		//TAB
		case  0x09: 
			if((current_symbol + 5) < (vga_row_count*vga_symbols_per_row))
				{
				//Add 5 Space
				str_array[current_symbol] = 0x0;
				current_symbol += 5;
			}
			break;
		//RETURN
		case  0x0D: 
			current_line = current_symbol / vga_symbols_per_row;
			if(current_line < vga_row_countm1)
				{
				str_array[current_symbol] = 0x0;
				current_symbol = (current_line+1)*vga_symbols_per_row;
				// efface ligne suivante
				i = vga_symbols_per_row;
				//Set pointers for clear string array
				ptr = &str_array[(current_line+1)*vga_symbols_per_row];
				while(i--)	*ptr++ = 0x0;
				}
				else	{
							str_array[current_symbol] = 0x0;
							current_symbol = 0x0;		// remet au dbut du tableau
							// efface la premire ligne
							i = vga_symbols_per_row;
							//Set pointers for clear string array
							ptr = &str_array[0];
							while(i--)	*ptr++ = 0x0;
							}
			break;
		default: str_array[current_symbol++] = received_symbol;
		}
}
//------------------------------------------------------------------------
//	Transmission over serial
void InitUART(void)
{
// Initialize the UART1 for BAUD = 19,200  
	U1MODE = 0x8400;	// enable + alternate pins
//	U1MODE = 0x8000;	// enable + normal pins
	U1STA = 0x0000;
	U1BRG = ((FCY/16)/BAUD) - 1;	// set baud to 19200
	IPC2bits.U1RXIP = 2;	// set RX interrupt priority level to 2 < 4 du Timer 
	IFS0bits.U1RXIF = 0;	// clear interrupt flag
	IEC0bits.U1RXIE = 1;			// enable RX interrupt

	SeqComm=SeqCommMax;
}
//------------------------------------------------------------------------
//Main routine
int main(void)
{
	setup_ports();
	InitVar();
	InitUART();
	initTimer();

	while(1)
		{		
		// tout se fait ds les ISR
		}	// end of while (1)
}	// end of main

//=============================================================================
//Error traps
//-----------------------------------------------------------------------------
//Oscillator Fail Error trap routine
void __attribute__((interrupt, auto_psv)) _ISR _OscillatorFail(void)
{	
	while(1); //Wait forever
}
//-----------------------------------------------------------------------------
//Address Error trap routine
void __attribute__((interrupt, auto_psv)) _ISR _AddressError(void)
{
	while(1); //Wait forever
}
//-----------------------------------------------------------------------------
//Stack Error trap routine
void __attribute__((interrupt, auto_psv)) _ISR _StackError(void)
{
	while(1); //Wait forever
}
//-----------------------------------------------------------------------------
//Math (Arithmetic) Error trap routine
void __attribute__((interrupt, auto_psv)) _ISR _MathError(void)
{
	while(1); //Wait forever
}
//---------------------------------------------------------------------
// This is a generic 1ms delay routine to give a 1mS to 65.5 Seconds delay
// For N = 1 the delay is 1 mS, for N = 65535 the delay is 65,535 mS. 
// Note that FCY is used in the computation.  Please make the necessary
// Changes(PLLx4 or PLLx8 etc) to compute the right FCY as in the define
// statement above.
//---------------------------------------------------------------------
void DelayNmSec(unsigned int N)
{
unsigned int j;
while(N--)
 	for(j=0;j < MILLISEC;j++);
}
//---------------------------------------------------------------------
